/* config.h.  Generated from config.h.in by configure.  */
/* config.h.in.  Generated from configure.ac by autoheader.  */

/* Define to 1 if you have the declaration of `isfinite', and to 0 if you
   don't. */
#define HAVE_DECL_ISFINITE 1

/* Define to 1 if you have the <dlfcn.h> header file. */
#define HAVE_DLFCN_H 1

/* Define to 1 if you have the `expm1' function. */
#define HAVE_EXPM1 1

/* Define to 1 if you have the `fabsl' function. */
#define HAVE_FABSL 1

/* Define to 1 if you have the `finite' function. */
#define HAVE_FINITE 1

/* Define to 1 if you have the `fmin' function. */
#define HAVE_FMIN 1

/* Define to 1 if you have the `ftruncate' function. */
#define HAVE_FTRUNCATE 1

/* Define to 1 if you have the GLPK library */
#define HAVE_GLPK 1

/* Define to 1 if you have the GMP library */
#define HAVE_GMP 1

/* Define to 1 if you have the <inttypes.h> header file. */
#define HAVE_INTTYPES_H 1

/* Define to 1 if you have the `isnan' function. */
#define HAVE_ISNAN 1

/* Define to 1 if you have the `arpack' library (-larpack). */
#define HAVE_LIBARPACK 1

/* Define to 1 if you have the `blas' library (-lblas). */
#define HAVE_LIBBLAS 1

/* Define to 1 if you have the `f2c' library (-lf2c). */
/* #undef HAVE_LIBF2C */

/* Define to 1 if you have the `lapack' library (-llapack). */
#define HAVE_LIBLAPACK 1

/* Define to 1 if you have the libxml2 libraries installed */
#define HAVE_LIBXML 1

/* Define to 1 if you have the `log1p' function. */
#define HAVE_LOG1P 1

/* Define to 1 if you have the `log2' function. */
#define HAVE_LOG2 1

/* Define to 1 if you have the <memory.h> header file. */
#define HAVE_MEMORY_H 1

/* Define to 1 if you have the `rint' function. */
#define HAVE_RINT 1

/* Define to 1 if you have the `rintf' function. */
#define HAVE_RINTF 1

/* Define to 1 if you have the `round' function. */
#define HAVE_ROUND 1

/* Define to 1 if you have the `snprintf' function. */
#define HAVE_SNPRINTF 1

/* Define to 1 if you have the <stdarg.h> header file. */
#define HAVE_STDARG_H 1

/* Define to 1 if you have the <stdint.h> header file. */
#define HAVE_STDINT_H 1

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the `stpcpy' function. */
#define HAVE_STPCPY 1

/* Define to 1 if the stpcpy function has a signature */
#define HAVE_STPCPY_SIGNATURE 1

/* Define to 1 if you have the `strcasecmp' function. */
#define HAVE_STRCASECMP 1

/* Define to 1 if you have the `strdup' function. */
#define HAVE_STRDUP 1

/* Define to 1 if you have the <strings.h> header file. */
#define HAVE_STRINGS_H 1

/* Define to 1 if you have the <string.h> header file. */
#define HAVE_STRING_H 1

/* Define to 1 if you have the <sys/int_types.h> header file. */
/* #undef HAVE_SYS_INT_TYPES_H */

/* Define to 1 if you have the <sys/stat.h> header file. */
#define HAVE_SYS_STAT_H 1

/* Define to 1 if you have the <sys/types.h> header file. */
#define HAVE_SYS_TYPES_H 1

/* Define to 1 if you have the sys/times.h header */
#define HAVE_TIMES_H 1

/* Define to 1 if you have the <time.h> header file. */
#define HAVE_TIME_H 1

/* Define to 1 if you want to use thread-local storage for global igraph
   structures */
/* #undef HAVE_TLS */

/* Define to 1 if you have the <unistd.h> header file. */
#define HAVE_UNISTD_H 1

/* Define to 1 if you have the `_strdup' function. */
/* #undef HAVE__STRDUP */

/* Keyword for thread local storage, or just static if not available */
#define IGRAPH_F77_SAVE static IGRAPH_THREAD_LOCAL

/* Keyword for thread local storage, or empty if not available */
#define IGRAPH_THREAD_LOCAL 

/* Define to 1 if you use the internal ARPACK library */
#define INTERNAL_ARPACK 1

/* Define to 1 if you use the internal BLAS library */
#undef INTERNAL_BLAS

/* Define to 1 if you use the internal F2C library */
#undef INTERNAL_F2C

/* Define to 1 if you use the internal GLPK library */
#undef INTERNAL_GLPK

/* Define to 1 if you use the internal LAPACK library */
#undef INTERNAL_LAPACK

/* Define to the sub-directory where libtool stores uninstalled libraries. */
#define LT_OBJDIR ".libs/"

/* Name of package */
#define PACKAGE "igraph"

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT "igraph@igraph.org"

/* Define to the full name of this package. */
#define PACKAGE_NAME "igraph"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "igraph 0.8.2"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "igraph"

/* Define to the home page for this package. */
#define PACKAGE_URL ""

/* Define to the version of this package. */
#define PACKAGE_VERSION "0.8.2"

/* Define to 1 if you have the ANSI C header files. */
#define STDC_HEADERS 1

/* If the compiler supports a TLS storage class define it to that here */
/* #undef TLS */

/* Version number of package */
#define VERSION "0.8.2"

/* Define to 1 if `lex' declares `yytext' as a `char *' by default, not a
   `char[]'. */
/* #undef YYTEXT_POINTER */
