/*################################################################################
  ##
  ##   Copyright (C) 2011-2020 Keith O'Hara
  ##
  ##   This file is part of the StatsLib C++ library.
  ##
  ##   Licensed under the Apache License, Version 2.0 (the "License");
  ##   you may not use this file except in compliance with the License.
  ##   You may obtain a copy of the License at
  ##
  ##       http://www.apache.org/licenses/LICENSE-2.0
  ##
  ##   Unless required by applicable law or agreed to in writing, software
  ##   distributed under the License is distributed on an "AS IS" BASIS,
  ##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  ##   See the License for the specific language governing permissions and
  ##   limitations under the License.
  ##
  ################################################################################*/

#ifndef _statslib_sanity_checks_HPP
#define _statslib_sanity_checks_HPP

#include "prob_val.hpp"

#include "bern.hpp"
#include "beta.hpp"
#include "binom.hpp"
#include "cauchy.hpp"
#include "chisq.hpp"
#include "exp.hpp"
#include "f.hpp"
#include "gamma.hpp"
#include "invgamma.hpp"
// #include "invwish.hpp"
#include "laplace.hpp"
#include "lnorm.hpp"
#include "logis.hpp"
// #include "mvnorm.hpp"
#include "norm.hpp"
#include "pois.hpp"
#include "t.hpp"
#include "unif.hpp"
#include "weibull.hpp"
// #include "wish.hpp"

#endif
